!MNH_LIC Copyright 1999-2023 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
!     ####################
      MODULE MODI_CONRESOL
!     ####################
!
INTERFACE
!
      SUBROUTINE CONRESOL(HLBCX,HLBCY,PDXX,PDYY,PDZX,PDZY,PDZZ,PRHODJ,PTHETAV, &
      PDXHATM,PDYHATM,PRHOM,PAF,PBF,PCF,PTRIGSX,PTRIGSY,KIFAXX,KIFAXY,         &
      KITR,PY,PPHI)
!  
IMPLICIT NONE
!
CHARACTER (LEN=4), DIMENSION(2), INTENT(IN) :: HLBCX    ! x-direction LBC type 
CHARACTER (LEN=4), DIMENSION(2), INTENT(IN) :: HLBCY    ! y-direction LBC type 
!
                                                   ! Metric coefficients:
REAL, DIMENSION(:,:,:), INTENT(IN)    :: PDXX      ! d*xx 
REAL, DIMENSION(:,:,:), INTENT(IN)    :: PDYY      ! d*yy 
REAL, DIMENSION(:,:,:), INTENT(INOUT) :: PDZX      ! d*zx 
REAL, DIMENSION(:,:,:), INTENT(INOUT) :: PDZY      ! d*zy 
REAL, DIMENSION(:,:,:), INTENT(IN)    :: PDZZ      ! d*zz
!
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PRHODJ    ! density of reference * J
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PTHETAV   ! virtual pot. temp. at time t
!
REAL, INTENT(IN) :: PDXHATM                     ! mean grid increment in the x
                                                ! direction
REAL, INTENT(IN) :: PDYHATM                     ! mean grid increment in the y
                                                ! direction
!
REAL, DIMENSION (:), INTENT(IN) :: PRHOM         !  XRHODJ mean on the X Y plane
                                                 !  localized at a mass level
!
REAL, DIMENSION(:), INTENT(IN)     :: PAF,PCF    ! vectors giving the nonvanishing
REAL, DIMENSION(:,:,:), INTENT(IN) :: PBF        ! elements of the tri-diag. 
                                                 ! matrix in the pressure eq.
!
                                                 ! arrays of sin or cos values
                                                 ! for the FFT :
REAL, DIMENSION(:), INTENT(IN) :: PTRIGSX        ! - along x
REAL, DIMENSION(:), INTENT(IN) :: PTRIGSY        ! - along y
!
                                                 ! decomposition in prime 
                                                 ! numbers for the FFT:
INTEGER, DIMENSION(19), INTENT(IN) :: KIFAXX      ! - along x
INTEGER, DIMENSION(19), INTENT(IN) :: KIFAXY      ! - along y
!
INTEGER, INTENT(IN) :: KITR                      ! number of iterations for the
                                                 ! pressure solver
!
REAL, DIMENSION(:,:,:), INTENT(IN) :: PY         ! RHS of the equation
!          
REAL, DIMENSION(:,:,:), INTENT(INOUT) :: PPHI    ! solution of the equation 
!
END SUBROUTINE CONRESOL
!
END INTERFACE
!
END  MODULE MODI_CONRESOL
!
!
!
!     #########################################################################
      SUBROUTINE CONRESOL(HLBCX,HLBCY,PDXX,PDYY,PDZX,PDZY,PDZZ,PRHODJ,PTHETAV, &
      PDXHATM,PDYHATM,PRHOM,PAF,PBF,PCF,PTRIGSX,PTRIGSY,KIFAXX,KIFAXY,         &
      KITR,PY,PPHI)
!     #########################################################################
!
!!****  *CONRESOL * - solve an elliptic equation by the conjugate residual
!!       method
!!
!!    PURPOSE
!!    -------
!       The purpose of this routine is to solve an elliptic equation using 
!     the preconditioned conjugate residual (CR) method. This is a version
!     of the scheme proposed by Skamarock, Smolarkiewicz and Klemp (MWR, 1997).
!     
!!**  METHOD
!!    ------
!!      The equation to be solved reads:
!!
!!             Q (PHI) = Y
!!
!!    where Q is the quasi-Laplacian ( subroutine QLAP ) and PHI the pressure
!!    function.
!!    We precondition the problem by the operator F :
!!             -1               -1 
!!            F   * Q (PHI) = F    (Y)
!!    F represents the flat Laplacian ie. without orography. Its inversion is 
!!    realized in the routine FLAT_INV. This equation is solved with a Conjugate
!!    Residual method.
!!    The initial guess is given by the pressure at the previous time step.
!!    The resolution stops after ITR iterations of the solver.
!!
!!    EXTERNAL
!!    --------
!!      Function QLAP: compute the complete quasi-Laplacian Q
!!      Subroutine FLAT_INV : invert the flat quasi-laplacien F
!!      Function DOTPROD: compute the dot product of 2 vectors
!!
!!    IMPLICIT ARGUMENTS
!!    ------------------ 
!!      Module MODI_QLAP: interface for the function QLAP
!!      Module MODI_FLAT_INV: interface for the subroutine FLAT_INV
!!      Module MODI_DOTPROD: interface for the function DOTPROD
!!
!!    REFERENCE
!!    ---------
!!      Book2 of documentation (routine CONRESOL)
!!      Skamarock, Smolarkiewicz and Klemp (1997) MWR
!!      
!!    AUTHOR
!!    ------
!!	J.-P. Pinty *Laboratoire d'Aerologie*
!!
!!    MODIFICATIONS
!!    -------------
!!      Original    25/08/99 
!!      J.-P. Pinty & P. Jabouille
!!                  11/07/00 bug in ZALPHA
!  P. Wautelet 22/11/2022: modernization of FFts (fixed-format F77 -> free-format F95)
!-------------------------------------------------------------------------------
!
!*       0.     DECLARATIONS
!               ------------
!
#ifdef MNH_OPENACC
USE MODE_MNH_ZWORK, ONLY: MNH_MEM_GET, MNH_MEM_POSITION_PIN, MNH_MEM_RELEASE
#endif
USE MODE_MPPDB
!
USE MODI_DOTPROD
USE MODI_FLAT_INV
USE MODI_QLAP
!
IMPLICIT NONE
!
!*      0.1    declarations of arguments
!
!
CHARACTER (LEN=4), DIMENSION(2), INTENT(IN) :: HLBCX    ! x-direction LBC type 
CHARACTER (LEN=4), DIMENSION(2), INTENT(IN) :: HLBCY    ! y-direction LBC type 
!
                                                   ! Metric coefficients:
REAL, DIMENSION(:,:,:), INTENT(IN)    :: PDXX      ! d*xx 
REAL, DIMENSION(:,:,:), INTENT(IN)    :: PDYY      ! d*yy 
REAL, DIMENSION(:,:,:), INTENT(INOUT) :: PDZX      ! d*zx 
REAL, DIMENSION(:,:,:), INTENT(INOUT) :: PDZY      ! d*zy 
REAL, DIMENSION(:,:,:), INTENT(IN)    :: PDZZ      ! d*zz
!
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PRHODJ    ! density of reference * J
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PTHETAV   ! virtual pot. temp. at time t
!
REAL, INTENT(IN) :: PDXHATM                     ! mean grid increment in the x
                                                ! direction
REAL, INTENT(IN) :: PDYHATM                     ! mean grid increment in the y
                                                ! direction
!
REAL, DIMENSION (:), INTENT(IN) :: PRHOM         !  mean of XRHODJ on the plane x y 
                                                 !  localized at a mass level
!
REAL, DIMENSION(:), INTENT(IN)     :: PAF,PCF    ! vectors giving the nonvanishing
REAL, DIMENSION(:,:,:), INTENT(IN) :: PBF        ! elements of the tri-diag. 
                                                 ! matrix in the pressure eq.
!
                                                 ! arrays of sin or cos values
                                                 ! for the FFT :
REAL, DIMENSION(:), INTENT(IN) :: PTRIGSX        ! - along x
REAL, DIMENSION(:), INTENT(IN) :: PTRIGSY        ! - along y
!
                                                 ! decomposition in prime 
                                                 ! numbers for the FFT:
INTEGER, DIMENSION(19), INTENT(IN) :: KIFAXX      ! - along x
INTEGER, DIMENSION(19), INTENT(IN) :: KIFAXY      ! - along y
!
INTEGER, INTENT(IN) :: KITR                      ! number of iterations for the
                                                 ! pressure solver
!
REAL, DIMENSION(:,:,:), INTENT(IN) :: PY         ! RHS of the equation
!          
REAL, DIMENSION(:,:,:), INTENT(INOUT) :: PPHI    ! solution of the equation 
!
!*      0.2    declarations of local variables
!
INTEGER :: JM                                    ! loop index   
!
#ifndef MNH_OPENACC
REAL, DIMENSION(SIZE(PPHI,1),SIZE(PPHI,2),SIZE(PPHI,3)) :: ZDELTA, ZKSI
     ! array containing the auxilary fields DELTA and KSI of the CR method
REAL, DIMENSION(SIZE(PPHI,1),SIZE(PPHI,2),SIZE(PPHI,3)) :: ZP, ZQ  
     ! array containing the auxilary fields P and Q of the CR method
REAL, DIMENSION(SIZE(PPHI,1),SIZE(PPHI,2),SIZE(PPHI,3)) :: ZRESIDUE
     ! array containing the error field at each iteration Q(PHI) - Y
#else
REAL, DIMENSION(:,:,:), pointer , contiguous :: ZDELTA, ZKSI ! array containing the auxilary fields DELTA and KSI of the CR method
REAL, DIMENSION(:,:,:), pointer , contiguous :: ZP, ZQ       ! array containing the auxilary fields P and Q of the CR method
REAL, DIMENSION(:,:,:), pointer , contiguous :: ZRESIDUE     ! array containing the error field at each iteration Q(PHI) - Y
#endif
!
REAL :: ZALPHA, ZLAMBDA      ! amplitude of the descent in the Conjugate
                             ! directions
REAL :: ZDOT_DELTA           ! dot product of ZDELTA by itself
#ifdef MNH_OPENACC
INTEGER  :: JIU,JJU,JKU
#endif
!
!-------------------------------------------------------------------------------

!$acc data present( PDXX, PDYY, PDZX, PDZY, PDZZ, PRHODJ, PTHETAV )              &
!$acc &    present( PRHOM, PAF, PBF, PCF, PTRIGSX, PTRIGSY, KIFAXX, KIFAXY, PY ) &
!$acc &    present( PPHI )

IF (MPPDB_INITIALIZED) THEN
  !Check all IN arrays
  CALL MPPDB_CHECK(PDXX,"CRESI beg:PDXX")
  CALL MPPDB_CHECK(PDYY,"CRESI beg:PDYY")
  CALL MPPDB_CHECK(PDZX,"CRESI beg:PDZX")
  CALL MPPDB_CHECK(PDZY,"CRESI beg:PDZY")
  CALL MPPDB_CHECK(PDZZ,"CRESI beg:PDZZ")
  CALL MPPDB_CHECK(PRHODJ,"CRESI beg:PRHODJ")
  CALL MPPDB_CHECK(PTHETAV,"CRESI beg:PTHETAV")
  CALL MPPDB_CHECK(PRHOM,"CRESI beg:PRHOM")
  CALL MPPDB_CHECK(PAF,"CRESI beg:PAF")
  CALL MPPDB_CHECK(PBF,"CRESI beg:PBF")
  CALL MPPDB_CHECK(PCF,"CRESI beg:PCF")
  CALL MPPDB_CHECK(PTRIGSX,"CRESI beg:PTRIGSX")
  CALL MPPDB_CHECK(PTRIGSY,"CRESI beg:PTRIGSY")
  CALL MPPDB_CHECK(KIFAXX,"CRESI beg:KIFAXX")
  CALL MPPDB_CHECK(KIFAXY,"CRESI beg:KIFAXY")
  CALL MPPDB_CHECK(PY,"CRESI beg:PY")
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PPHI,"CRESI beg:PPHI")
END IF
!
!*       1.    INITIALIZATIONS
!              ---------------
!
#ifdef MNH_OPENACC
JIU =  size(PPHI, 1 )
JJU =  size(PPHI, 2 )
JKU =  size(PPHI, 3 )

!Pin positions in the pools of MNH memory
CALL MNH_MEM_POSITION_PIN( 'CONRESOL' )

CALL MNH_MEM_GET( ZDELTA,   JIU, JJU, JKU )
CALL MNH_MEM_GET( ZKSI,     JIU, JJU, JKU )
CALL MNH_MEM_GET( ZP,       JIU, JJU, JKU )
CALL MNH_MEM_GET( ZQ,       JIU, JJU, JKU )
CALL MNH_MEM_GET( ZRESIDUE, JIU, JJU, JKU )
#endif

!
!*       1.1    compute the vector: r^(0) =  Q(PHI) - Y
!    
CALL QLAP(ZRESIDUE,HLBCX,HLBCY,PDXX,PDYY,PDZX,PDZY,PDZZ,PRHODJ,PTHETAV,PPHI)
!$acc kernels present( PY, ZRESIDUE )
ZRESIDUE(:,:,:) = ZRESIDUE(:,:,:) - PY(:,:,:)
!$acc end kernels
!
!*       1.2    compute the vector: p^(0) = F^(-1)*( Q(PHI) - Y )
!    
CALL FLAT_INV(HLBCX,HLBCY,PDXHATM,PDYHATM,PRHOM,PAF,PBF,PCF,  &
                     PTRIGSX,PTRIGSY,KIFAXX,KIFAXY,ZRESIDUE,ZP)
!
!*       1.3    compute the vector: delta^(0) = Q ( p^(0) )
!
CALL QLAP(ZDELTA,HLBCX,HLBCY,PDXX,PDYY,PDZX,PDZY,PDZZ,PRHODJ,PTHETAV,ZP)
!
!-------------------------------------------------------------------------------
!
!*       2.    ITERATIVE LOOP
!              --------------
!
DO JM = 1,KITR                                                         
!
!*       2.1    compute the step LAMBDA
!
  ZDOT_DELTA = DOTPROD(ZDELTA,  ZDELTA,HLBCX,HLBCY)            ! norm of DELTA
  ZLAMBDA  = - DOTPROD(ZRESIDUE,ZDELTA,HLBCX,HLBCY) / ZDOT_DELTA
!
!*       2.2    update the pressure function PHI
!
!$acc kernels
  PPHI(:,:,:) = PPHI(:,:,:) + ZLAMBDA * ZP(:,:,:)
!$acc end kernels
!
!
  IF( JM == KITR ) EXIT
!
!
!*       2.3    update the residual error: r
!
!$acc kernels
  ZRESIDUE(:,:,:) = ZRESIDUE(:,:,:) + ZLAMBDA * ZDELTA(:,:,:)
!$acc end kernels
!                        
!*       2.4    compute the vector: q = F^(-1)*( Q(PHI) - Y )
!    
  CALL FLAT_INV(HLBCX,HLBCY,PDXHATM,PDYHATM,PRHOM,PAF,PBF,PCF,  &
                       PTRIGSX,PTRIGSY,KIFAXX,KIFAXY,ZRESIDUE,ZQ)
!   
!*       2.5     compute the auxiliary field: ksi = Q ( q )
!
  CALL QLAP(ZKSI,HLBCX,HLBCY,PDXX,PDYY,PDZX,PDZY,PDZZ,PRHODJ,PTHETAV,ZQ)
!                                         -1
!*       2.6     compute the step ALPHA
!  
  ZALPHA = - DOTPROD(ZKSI,ZDELTA,HLBCX,HLBCY) / ZDOT_DELTA   ! lambda
!
!*       2.7     update p and DELTA
!
!$acc kernels
  ZP(:,:,:)     = ZQ(:,:,:)   + ZALPHA * ZP(:,:,:)
  ZDELTA(:,:,:) = ZKSI(:,:,:) + ZALPHA * ZDELTA(:,:,:)
!$acc end kernels
!
END DO              ! end of the loop for the iterative solver
!
#ifdef MNH_OPENACC
!Release all memory allocated with MNH_MEM_GET calls since last call to MNH_MEM_POSITION_PIN
CALL MNH_MEM_RELEASE( 'CONRESOL' )
#endif

IF (MPPDB_INITIALIZED) THEN
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PPHI,"CONRESOL end:PPHI")
END IF

!$acc end data

!-------------------------------------------------------------------------------
!
END SUBROUTINE CONRESOL
